/*
 * iws statistics interface
 */

/* this extension is multi-threaded */
#define _REENTRANT 1

#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pthread.h>
#include "top.h"
#include "utils.h"

#define HTTP_PORT 80
#define AGGIE_PORT 24443

static char PERFSTATS[] = "GET /perfdump HTTP/1.0\r\n\r\n";
#define SIZEOF_PERFSTATS (sizeof(PERFSTATS)-1)

static char TAG[] = "\nTotal number of requests:";
#define SIZEOF_TAG (sizeof(TAG)-1)

#ifndef MILLISEC
#define MILLISEC 1000
#endif

/* these are the protected variables */
static pthread_mutex_t datalock;
static pthread_cond_t wakeup;
static int hits = -1;
static int errors = -1;
static int stalecnt = 0;

/* these variables are only used by the background thread */
static unsigned long lasttotal = 0;
static unsigned long currtotal = 0;
static struct timeval lasttime = { 0, 0 };
static struct timeval currtime = { 0, 0 };
static char *null = NULL;


int
get_iws_errors()

{
    struct sockaddr_in in_sock;
    int s;
    int cnt;
    int rdbufsiz;
    int answer;
    char rdbuf[4096];
    char *p;
    char *lasts;

    /* clear the socket structure */
    memset(&in_sock, 0, sizeof(in_sock));

    /* initialize socket to use loopback */
    in_sock.sin_family = AF_INET;
    in_sock.sin_addr.s_addr = INADDR_LOOPBACK;
    in_sock.sin_port = AGGIE_PORT;

    /* create socket */
    if ((s = socket(PF_INET, SOCK_STREAM, 0)) == -1)
    {
	return -1;
    }

    /* connect */
    if (connect(s, (struct sockaddr *)(&in_sock), sizeof(in_sock)) == -1)
    {
	close(s);
	return -1;
    }

    /* read to eof */
    p = rdbuf;
    rdbufsiz = sizeof(rdbuf) - 1;
    while (rdbufsiz > 0 && (cnt = read(s, p, rdbufsiz)) > 0)
    {
	rdbufsiz -= cnt;
	p += cnt;
    }

    /* null terminate the buffer */
    *p = '\0';

    /* close the connection */
    close(s);

    /* walk thru the buffer looking for the count */
    answer = -1;
    if ((p = strstr(rdbuf, "\nli")) != NULL)
    {
	/* find the right number */
	p = strtok_r(p+1, " \t", &lasts);      /* type = li */
	p = strtok_r(NULL, " \t", &lasts);     /* name */
	p = strtok_r(NULL, " \t", &lasts);     /* port */
	p = strtok_r(NULL, " \t", &lasts);     /* hits */
	p = strtok_r(NULL, " \t", &lasts);     /* errors */
	answer = atoi(p);
    }

    return answer;
}

int
get_iws_hits()

{
    struct sockaddr_in in_sock;
    int s;
    int cnt;
    int rdbufsiz;
    long difftotal;
    long diffmsec;
    long answer;
    char rdbuf[4096];
    char *p;

    /* clear the socket structure */
    memset(&in_sock, 0, sizeof(in_sock));

    /* initialize socket to use loopback */
    in_sock.sin_family = AF_INET;
    in_sock.sin_addr.s_addr = INADDR_LOOPBACK;
    in_sock.sin_port = HTTP_PORT;

    /* create socket */
    if ((s = socket(PF_INET, SOCK_STREAM, 0)) == -1)
    {
	return -1;
    }

    /* connect */
    if (connect(s, (struct sockaddr *)(&in_sock), sizeof(in_sock)) == -1)
    {
	return -1;
    }

    /* send request */
    write(s, PERFSTATS, SIZEOF_PERFSTATS);

    /* read to eof */
    p = rdbuf;
    rdbufsiz = sizeof(rdbuf) - 1;
    while (rdbufsiz > 0 && (cnt = read(s, p, rdbufsiz)) > 0)
    {
	rdbufsiz -= cnt;
	p += cnt;
    }

    /* null terminate the buffer */
    *p = '\0';

    /* close the connection */
    close(s);

    /* stamp the time */
    gettimeofday(&currtime, &null);

    /* walk thru the buffer looking for the count */
    if ((p = strstr(rdbuf, TAG)) != NULL)
    {
	/* find the number */
	p += SIZEOF_TAG;
	p += strspn(p, " \t");
	currtotal = atoi(p);
    }

    /* compare to the last total */
    if (currtotal < lasttotal)
    {
	/* we've had a reset */
	lasttotal = 0;
    }
    difftotal = currtotal - lasttotal;

    /* we only do the rest if the last time stamp was valid */
    if (lasttime.tv_sec != 0)
    {
	/* compare time stamps */
	diffmsec = (currtime.tv_usec - lasttime.tv_usec) / 1000 +
	    ((currtime.tv_sec - lasttime.tv_sec) * MILLISEC);

	/* extrapolate difftotal to a per-minute measurement */
	answer = (difftotal * MILLISEC * 60 + 500) / diffmsec;
    }
    else
    {
	answer = 0;
    }

    /* remember current figures for next time */
    lasttime = currtime;
    lasttotal = currtotal;

    /* that's it */
    return (int)answer;
}

static void *
iws_thread(void *arg)

{
    int h;
    int e;

    while(1)
    {
	/* collect the information */
	h = get_iws_hits();
	e = get_iws_errors();

	/* lock and copy the information out */
	pthread_mutex_lock(&datalock);
	hits = h;
	errors = e;
	stalecnt = 0;

	/* wait to be triggered for the next round */
	pthread_cond_wait(&wakeup, &datalock);
	pthread_mutex_unlock(&datalock);
    }
}

static int
cnn_minibar(char *buff, int max)

{
    int r = 0;
    int h;
    int e;
    int stale;
    char hitbuf[16];
    char errbuf[16];

    /* get the global data */
    pthread_mutex_lock(&datalock);
    h = hits;
    e = errors;
    stale = stalecnt++;
    pthread_cond_signal(&wakeup);
    pthread_mutex_unlock(&datalock);

    /* preformat hits */
    if (h < 0)
    {
	strcpy(hitbuf, "DOWN");
    }
    else if (stale >= 3)
    {
	strcpy(hitbuf, "TIMEOUT");
    }
    else
    {
	sprintf(hitbuf, "%d", h);
    }

    /* preformat errors */
    errbuf[0] = '\0';
    if (e >= 0)
    {
	sprintf(errbuf, "/%d", e);
    }

    /* put it all together */
    r = snprintf(buff, max, "WS: %s%s", hitbuf, errbuf);

    return r;
}

int
init_ext(struct ext_decl *ed)

{
    struct stat st;
    pthread_attr_t attr;
    pthread_t thr;

    if (stat("/opt/ns", &st) == -1)
    {
	return 0;
    }

    /* initialize the pthreads system */
    pthread_attr_init(&attr);
    pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

    /* initialize the pthread data structures */
    pthread_mutex_init(&datalock, NULL);
    pthread_cond_init(&wakeup, NULL);

    /* start the background thread */
    pthread_create(&thr, &attr, iws_thread, NULL);

    /* fill in the extension info */
    ed->f_minibar = cnn_minibar;
    ed->f_display = NULL;

    return 1;
}

