/*
 * top - a top users display for Unix
 *
 * SYNOPSIS:  PowerPC running AIX 5 or higher with 32-bit kernel
 *
 * DESCRIPTION:
 * This is the machine-dependent module for AIX 5 and higher
 * with 32-bit kernel.  Use aix5_64 for 64-bit kernel.
 * It is currenlty only tested on PowerPC architectures.
 *
 * TERMCAP: -lcurses
 *
 * CFLAGS: -DORDER -DHAVE_GETOPT
 *
 * LIBS: 
 *
 * AUTHOR:  Jonathan Chen <jon+top@spock.org>
 *          based on aix4 module by Joep Vesseur <joep@fwi.uva.nl>
 *
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <fcntl.h>
#include <nlist.h>
#include <sys/sysinfo.h>
#include <procinfo.h>
#include <sys/proc.h>
#include <sys/times.h>
#include <sys/param.h>
#include <sys/systemcfg.h>
#include <sys/vminfo.h>
#include <time.h>
#include <pwd.h>
#include "top.h"
#include "machine.h"
#include "utils.h"

#ifdef __64BIT__
# define loadavg_t uint64_t
# define sysinfo sysinfo64
#else
# define loadavg_t uint32_t
#endif

#define PROCRESS(p) (((p)->pi_trss + (p)->pi_drss)*4)
#define PROCSIZE(p) (((p)->pi_tsize/1024+(p)->pi_dvm)*4)
#define PROCTIME(pi) (pi->pi_ru.ru_utime.tv_sec + pi->pi_ru.ru_stime.tv_sec)
#define GETNICE(n) ((n) & SFIXPRI) ? (P_NICE_MAX+1) : ((int)(n) - 20)
#define pi_Xcpu pi_resvd[0]
#define ncpus _system_configuration.ncpus

#define KMEM "/dev/kmem"

/* Indices in the nlist array */
#define X_AVENRUN       0
#define X_SYSINFO       1

static struct nlist nlst[] = {
    { "avenrun", 0, 0, 0, 0, 0 }, /* 0 */
    { "sysinfo", 0, 0, 0, 0, 0 }, /* 1 */
    {  NULL, 0, 0, 0, 0, 0 }
};


/* get_process_info returns handle. definition is here */
struct handle {
	struct procsinfo **next_proc;
	int remaining;
};

static struct handle handle;

/*
 *  These definitions control the format of the per-process area
 */

static char header[] =
  "   PID X        PRI NICE   SIZE   RES STATE   TIME   WCPU    CPU COMMAND";
/* 0123456   -- field to fill in starts at header+6 */
#define UNAME_START 7

#define Proc_format \
	"%6ld %-8.8s %3d %4d %5d%c %4d%c %-5s %6s %5.2f%% %5.2f%% %.14s%s"

/* these are for detailing the process states */
static int process_states[9];
static char *procstatenames[] = {
    " none, ", " sleeping, ", " state2, ", " runnable, ",
    " idle, ", " zombie, ", " stopped, ", " running, ", " swapped, ",
    NULL
};


/* these are for detailing the cpu states */
static int cpu_states[4];
static char *cpustatenames[] = {
    "idle", "user", "kernel", "wait",
    NULL
};

/* these are for detailing the memory statistics */
static int memory_stats[7];
static char *memorynames[] = {
    "M Total, ", "M Used, ", "M Virtual, ", "M Pinned, ", "M Free ", NULL
};
#define M_REAL 	   0
#define M_USED     1
#define M_VIRTUAL  2
#define M_PIN      3
#define M_FREE     4

static char *state_abbrev[] = {
    "", "sleep", "", "", "sleep", "zomb", "stop", "run", "swap"
};

int compare_size(struct procsinfo **ppi1, struct procsinfo **ppi2);
int compare_prio(struct procsinfo **ppi1, struct procsinfo **ppi2);
int compare_time(struct procsinfo **ppi1, struct procsinfo **ppi2);
int compare_res(struct procsinfo **ppi1, struct procsinfo **ppi2);
int compare_cpu(struct procsinfo **ppi1, struct procsinfo **ppi2);
int getkval(unsigned long offset, caddr_t ptr, int size, char *refstr);
void quit();

/* sorting orders. first is default */
static char *ordernames[] = {
    "cpu", "size", "res", "time", "pri", NULL
};

int (*proc_compares[])() = {
    compare_cpu,
    compare_size,
    compare_res,
    compare_time,
    compare_prio,
    NULL
};

/* useful globals */
static int kmem;			/* file descriptor */

/* offsets in kernel */
static unsigned long avenrun_offset;
static unsigned long sysinfo_offset;

/* the runqueue length is a cumulative value. keep old value */
long old_runque;

/* process info */
static int nprocs;			/* maximum nr of procs in proctab */
static int nthrds;
static struct procsinfo *p_info;	/* needed for vm and ru info */
static struct thrdsinfo *t_info;	/* needed for vm and ru info */
static struct procsinfo **pref;	/* processes selected for display */
static int pref_len;			/* number of processes selected */

/* needed to calculate WCPU */
static time_t curtime;


/*
 * Initialize globals, get kernel offsets and stuff...
 */
int machine_init(struct statics *statics) {
    time_t uptime, now;
    struct tms tbuf;

    if ((kmem = open(KMEM, O_RDONLY)) == -1) {
	perror(KMEM);
	return -1;
    }

    /* get kernel symbol offsets */
    if (knlist(nlst, 2, sizeof(struct nlist)) != 0) {
	perror("knlist");
	return -1;
    }
    avenrun_offset = nlst[X_AVENRUN].n_value;
    sysinfo_offset = nlst[X_SYSINFO].n_value;

    nprocs = PROCMASK(PIDMAX);
    if (nprocs > 1024) nprocs = 1024;
    nthrds = 256;
    p_info = (struct procsinfo *)malloc(nprocs * sizeof (struct procsinfo));
    t_info = (struct thrdsinfo *)malloc(nthrds * sizeof (struct thrdsinfo));
    pref = (struct procsinfo **)malloc(nprocs * sizeof (struct procsinfo *));

    if (!t_info || !p_info || !pref) {
	fprintf(stderr, "top: not enough memory\n");
	return -1;
    }

    /* set boot time */
    now = time(NULL);
    uptime = times(&tbuf) / HZ;
    statics->boottime = now - uptime;

    statics->procstate_names = procstatenames;
    statics->cpustate_names = cpustatenames;
    statics->memory_names = memorynames;
    statics->order_names = ordernames;
    statics->flags.idle = 1;

    return(0);
}



char *format_header(char *uname_field) {
    char *ptr;

    ptr = header + UNAME_START;
    while (*uname_field != '\0')
    {
	*ptr++ = *uname_field++;
    }

    return(header);
}




void get_system_info(struct system_info *si) {
    loadavg_t load_avg[3];
    struct sysinfo s_info;
    int i;
    double total = 0;
    struct vminfo vminfo;
    /* used for calculating cpu state percentages */
#ifdef __64BIT__
    static long long cp_time[CPU_NTIMES];
    static long long cp_old[CPU_NTIMES];
    static long long cp_diff[CPU_NTIMES];
#else
    static long cp_time[CPU_NTIMES];
    static long cp_old[CPU_NTIMES];
    static long cp_diff[CPU_NTIMES];
#endif


    /* get the load avarage array */
    getkval(avenrun_offset, (caddr_t)load_avg, sizeof load_avg, "avenrun");

    /* get the sysinfo structure */
    getkval(sysinfo_offset, (caddr_t)&s_info, sizeof s_info, "sysinfo");

    /* convert load avarages to doubles */
    for (i = 0; i < 3; i++)
	si->load_avg[i] = (double)load_avg[i]/65536.0;

    /* calculate cpu state in percentages */
    for (i = 0; i < CPU_NTIMES; i++) {
	cp_old[i] = cp_time[i];
	cp_time[i] = s_info.cpu[i];
	cp_diff[i] = cp_time[i] - cp_old[i];
	total += cp_diff[i];
    }

    if (total > 5) { /* rate limit, so cpu_states doesn't get negative */
        total = total/1000.0;  /* top itself will correct this */
        for (i = 0; i < CPU_NTIMES; i++) {
            cpu_states[i] = cp_diff[i] / total;
        }
    }

    /* calculate memory statistics, scale 4K pages to megabytes */
    vmgetinfo(&vminfo, VMINFO, sizeof(struct vminfo));
#define PAGE_TO_MB(a) ((a)*4/1024)
    memory_stats[M_REAL]     = PAGE_TO_MB(vminfo.memsizepgs);
    memory_stats[M_USED]     = PAGE_TO_MB(vminfo.memsizepgs - vminfo.numfrb);
    memory_stats[M_VIRTUAL]  = PAGE_TO_MB(vminfo.numvpages);
    memory_stats[M_PIN]     = PAGE_TO_MB(vminfo.numwsegpin);
    memory_stats[M_FREE]     = PAGE_TO_MB(vminfo.numfrb);

    /* runnable processes */
    process_states[0] = s_info.runque - old_runque;
    old_runque = s_info.runque;

    si->cpustates = cpu_states;
    si->memory = memory_stats;
}

caddr_t get_process_info(struct system_info *si, struct process_select *sel, int compare_index) {
    int i, nproc;
    int active_procs = 0, total_procs = 0;
    struct procsinfo *pp, **p_pref = pref;
    pid_t procsindex = 0;
    tid_t thrindex;

    si->procstates = process_states;

    curtime = time(0);

    /* get the procsinfo structures of all running processes */
    nproc = getprocs(p_info, sizeof (struct procsinfo), NULL, 0, 
		     &procsindex, nprocs);
    if (nproc < 0) {
	perror("getprocs");
	quit(1);
    }

    /* the swapper has no cmd-line attached */
    strcpy(p_info[0].pi_comm, "swapper");
    
    memset(process_states, 0, sizeof process_states);

    /* build a list of pointers to processes to show. walk through the
     * list of procsinfo structures instead of the proc table since the
     * mapping of procsinfo -> proctable is easy, the other way around
     * is cumbersome
     */
    for (pp = p_info, i = 0; i < nproc; pp++, i++) {
	int tcount, j;

	thrindex=0;
	tcount = getthrds(pp->pi_pid, t_info, sizeof(struct thrdsinfo), &thrindex, nthrds);
	pp->pi_Xcpu = 0;
	for (j = 0; j < tcount; j++)
	  pp->pi_Xcpu += t_info[j].ti_cpu;

	/* AIX marks all runnable processes as ACTIVE. We want to know
	   which processes are sleeping, so check used cpu ticks and adjust
	   status field accordingly
	 */
	if (pp->pi_state == SACTIVE && pp->pi_Xcpu == 0)
	    pp->pi_state = SIDL;

        if (pp->pi_state && (sel->system || ((pp->pi_flags & SKPROC) == 0))) {
	    total_procs++;
	    process_states[pp->pi_state]++;
	    if ( (pp->pi_state != SZOMB) &&
		(sel->idle || pp->pi_Xcpu != 0 || (pp->pi_state == SACTIVE))
		&& (sel->uid == -1 || pp->pi_uid == (uid_t)sel->uid)) {
                *p_pref++ = pp;
		active_procs++;
	    }
	}
    }   

    /* the pref array now holds pointers to the procsinfo structures in
     * the p_info array that were selected for display
     */

    /* sort if requested */
    if (si->p_active)
	qsort((char *)pref, active_procs, sizeof (struct procsinfo *), 
	      proc_compares[compare_index]);
    
    si->last_pid = -1;		/* no way to figure out last used pid */
    si->p_total = total_procs;
    si->p_active = pref_len = active_procs;

    handle.next_proc = pref;
    handle.remaining = active_procs;

    return((caddr_t)&handle);
}


/* define what weighted cpu is. use definition of %CPU from 'man ps(1)' */
#define weighted_cpu(pp) (PROCTIME(pp) == 0 ? 0.0 : \
                        (((PROCTIME(pp)*100.0)/(curtime-pi->pi_start)/ncpus)))

char *format_next_process(caddr_t handle, char *(*get_userid)()) {
    struct handle *hp;
    struct procsinfo *pi;
    long cpu_time;
    int proc_size, proc_ress;
    char size_unit = 'K';
    char ress_unit = 'K';
    static char fmt[128];		/* static area where result is built */

    hp = (struct handle *)handle;
    if (hp->remaining == 0) {	/* safe guard */
	fmt[0] = '\0';
	return fmt;
    }
    pi = *(hp->next_proc++);
    hp->remaining--;

    cpu_time = PROCTIME(pi);

    /* we disply sizes up to 10M in KiloBytes, beyond 10M in MegaBytes */
    if ((proc_size = (pi->pi_tsize/1024+pi->pi_dvm)*4) > 10240) {
	proc_size /= 1024;
	size_unit = 'M';
    }
    if ((proc_ress = (pi->pi_trss + pi->pi_drss)*4) > 10240) {
	proc_ress /= 1024;
	ress_unit = 'M';
    }

    sprintf(fmt, Proc_format ,
            pi->pi_pid,					  /* PID */
            (*get_userid)(pi->pi_uid),			  /* login name */
            getpriority(PRIO_PROCESS, pi->pi_pid),
	    GETNICE(pi->pi_nice),			  /* fixed or vari */
            proc_size,					  /* size */
            size_unit,					  /* K or M */
            proc_ress,					  /* resident */
            ress_unit,					  /* K or M */
            state_abbrev[pi->pi_state],			  /* process state */
            format_time(cpu_time),			  /* time used */
	    weighted_cpu(pi),	                          /* WCPU */
	    (double)pi->pi_Xcpu/1.2,                      /* CPU */
            printable(pi->pi_comm),                       /* COMM */
	    (pi->pi_flags & SKPROC) == 0 ? "" : " (sys)"  /* kernel process? */
	    );
    return(fmt);
}


/*
 *  getkval(offset, ptr, size, refstr) - get a value out of the kernel.
 *	"offset" is the byte offset into the kernel for the desired value,
 *  	"ptr" points to a buffer into which the value is retrieved,
 *  	"size" is the size of the buffer (and the object to retrieve),
 *  	"refstr" is a reference string used when printing error meessages,
 *	    if "refstr" starts with a '!', then a failure on read will not
 *  	    be fatal (this may seem like a silly way to do things, but I
 *  	    really didn't want the overhead of another argument).
 *  	
 */
int getkval(unsigned long offset, caddr_t ptr, int size, char *refstr) {
    int upper_2gb = 0;

    /* reads above 2Gb are done by seeking to offset%2Gb, and supplying
     * 1 (opposed to 0) as fourth parameter to readx (see 'man kmem')
     */
    if (offset > 1<<31) {
	upper_2gb = 1;
	offset &= 0x7fffffff;
    }

    if (lseek(kmem, offset, SEEK_SET) != offset) {
	fprintf(stderr, "top: lseek: %s: %s\n", refstr, strerror(errno));
	quit(2);
    }

    if (readx(kmem, ptr, size, upper_2gb) != size) {
	if (*refstr == '!')
	    return 0;
	else {
	    fprintf(stderr, "top: kvm_read: %s: %s\n", refstr, strerror(errno));
	    quit(2);
	}
    }

    return 1 ;
}
    
/* comparison routine for qsort */
/*
 * The following code is taken from the solaris module and adjusted
 * for AIX -- JV .
 */

#define ORDERKEY_PCTCPU \
           if (lresult = pi2->pi_Xcpu - pi1->pi_Xcpu, \
               (result = lresult > 0 ? 1 : lresult < 0 ? -1 : 0) == 0)

#define ORDERKEY_CPTICKS \
           if ((result = PROCTIME(pi2) - PROCTIME(pi1)) == 0)


#define ORDERKEY_STATE \
           if ((result = sorted_state[pi2->pi_state]  \
                         - sorted_state[pi1->pi_state])  == 0)

/* Nice values directly reflect the process' priority, and are always >0 ;-) */
#define ORDERKEY_PRIO \
	   if ((result = GETNICE(pi1->pi_nice) - GETNICE(pi2->pi_nice)) == 0) 

#define ORDERKEY_RSSIZE \
           if ((result = PROCRESS(pi2) - PROCRESS(pi1)) == 0)
#define ORDERKEY_MEM \
           if ((result = PROCSIZE(pi2) - PROCSIZE(pi1)) == 0)

static unsigned char sorted_state[] =
{
    0, /* not used */
    0,
    0,
    0,
    3,                          /* sleep */
    1,				/* zombie */
    4,				/* stop */
    6,				/* run */
    2,				/* swap */
};

/* compare_cpu - the comparison function for sorting by cpu percentage */

int compare_cpu(struct procsinfo **ppi1, struct procsinfo **ppi2) {
    struct procsinfo *pi1 = *ppi1, *pi2 = *ppi2;
    int result;
    long lresult;

    ORDERKEY_PCTCPU
    ORDERKEY_CPTICKS
    ORDERKEY_STATE
    ORDERKEY_PRIO
    ORDERKEY_RSSIZE
    ORDERKEY_MEM
    ;

    return result;
}
    

/* compare_size - the comparison function for sorting by total memory usage */

int compare_size(struct procsinfo **ppi1, struct procsinfo **ppi2) {
    struct procsinfo *pi1 = *ppi1, *pi2 = *ppi2;
    int result;
    long lresult;

    ORDERKEY_MEM
    ORDERKEY_RSSIZE
    ORDERKEY_PCTCPU
    ORDERKEY_CPTICKS
    ORDERKEY_STATE
    ORDERKEY_PRIO
    ;

    return result;
}
    

/* compare_res - the comparison function for sorting by resident set size */

int compare_res(struct procsinfo **ppi1, struct procsinfo **ppi2) {
    struct procsinfo *pi1 = *ppi1, *pi2 = *ppi2;
    int result;
    long lresult;

    ORDERKEY_RSSIZE
    ORDERKEY_MEM
    ORDERKEY_PCTCPU
    ORDERKEY_CPTICKS
    ORDERKEY_STATE
    ORDERKEY_PRIO
    ;

    return result;
}
    

/* compare_time - the comparison function for sorting by total cpu time */

int compare_time(struct procsinfo **ppi1, struct procsinfo **ppi2) {
    struct procsinfo *pi1 = *ppi1, *pi2 = *ppi2;
    int result;
    long lresult;

    ORDERKEY_CPTICKS
    ORDERKEY_PCTCPU
    ORDERKEY_STATE
    ORDERKEY_PRIO
    ORDERKEY_MEM
    ORDERKEY_RSSIZE
    ;

    return result;
}
    

/* compare_prio - the comparison function for sorting by cpu percentage */

int compare_prio(struct procsinfo **ppi1, struct procsinfo **ppi2) {
    struct procsinfo *pi1 = *ppi1, *pi2 = *ppi2;
    int result;
    long lresult;

    ORDERKEY_PRIO
    ORDERKEY_PCTCPU
    ORDERKEY_CPTICKS
    ORDERKEY_STATE
    ORDERKEY_RSSIZE
    ORDERKEY_MEM
    ;

    return result;
}
    

int proc_owner(int pid) {
   struct procsinfo **prefp = pref;
   int cnt = pref_len;

   while (--cnt >= 0) {
       if ((*prefp)->pi_pid == pid)
	   return (*prefp)->pi_uid;
       prefp++;
   }
   
   return(-1);
}


